# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: Copyright The Geneva Authors

import argparse

from textual_serve.server import Server

from geneva.tui.main import GenevaLake


def server() -> None:
    Server(f"python -m {__name__}.main").serve(debug=True)


def run_console(args: argparse.Namespace) -> None:
    if args.server:
        server()
    else:
        GenevaLake().run()


def register_console_parser(subparsers: argparse._SubParsersAction) -> None:
    parser = subparsers.add_parser("console", description="Run a textual server")

    parser.add_argument("--server", action="store_true", help="Run a textual server")

    parser.set_defaults(func=run_console)



# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: Copyright The Geneva Authors

import json

import cloudpickle
from textual.app import ComposeResult
from textual.containers import Horizontal, Vertical, VerticalScroll
from textual.widgets import Button, DataTable, Footer, Header, Static

from geneva.apply import ReadTask
from geneva.debug.entrypoint import run_udf_task_with_debug
from geneva.debug.logger import CheckpointStoreErrorLogger
from geneva.tui._screen import DismissOnQuitScreen
from geneva.tui.error import show_error


class ErrorLogsScreen(DismissOnQuitScreen):
    def __init__(
        self,
        *args,
        error_logger: CheckpointStoreErrorLogger,
        **kwargs,
    ) -> None:
        super().__init__(*args, **kwargs)
        self.title = "Error Logs"
        self.error_logger = error_logger

    def compose(self) -> ComposeResult:
        yield Header(name=self.title)
        yield Footer()
        message = ""
        buttons = Vertical(
            Button("Connect to Debugger", id="debug"),
            Button("Run in Profiler", id="profile"),
            id="buttons",
        )

        self.table = get_error_table(self.error_logger)
        self.table.id = "error_table"

        yield Vertical(
            self.table,
            Horizontal(
                buttons,
                VerticalScroll(Static(message)),
            ),
        )

    def on_button_pressed(self, event: Button.Pressed) -> None:
        """An action to select the cursor."""

        row = self.table.cursor_row
        row = self.table.get_row_at(row)
        key = row[0]

        try:
            error_row = self.error_logger.get_error_row(key)
            task_json = error_row["task"][0].as_py()
            udf_data = error_row["udf"][0].as_py()

            task_dict = json.loads(task_json)
            task = ReadTask(**task_dict)
            udf = cloudpickle.loads(udf_data)
        except Exception as e:
            show_error(self.app, ex=e)
            return

        with self.app.suspend():
            run_udf_task_with_debug(task, udf)

    CSS = """
    #buttons {
        width: 10vw;
    }
    #error_table {
        height: 80vh;
        width: 100vw;
    }
    """


def get_error_table(
    error_logger: CheckpointStoreErrorLogger,
) -> DataTable:
    error_keys = list(error_logger.list_errors())

    table = DataTable(cursor_type="row")
    table.add_columns(
        "ID",
        "URI",
        "Input Columns",
        "Frament ID",
        "Offset",
        "Limit",
        "Batch Size",
        "Error",
    )

    for key in error_keys:
        error_row = error_logger.get_error_row(key)
        task_json = error_row["task"][0].as_py()
        task_dict = json.loads(task_json)
        error = error_row["error"][0].as_py()

        table.add_row(
            key,
            task_dict["uri"],
            task_dict["columns"],
            task_dict["frag_id"],
            task_dict["offset"],
            task_dict["limit"],
            task_dict["batch_size"],
            error,
        )

    return table



# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: Copyright The Geneva Authors

from pathlib import Path

import lance
import pyarrow as pa

from geneva import LanceCheckpointStore, udf
from geneva.runners.dataflow.dataflow import DataflowOptions, run_dataflow_add_column


@udf(data_type=pa.int32(), batch_size=8)
def add_one(a) -> int:
    return a + 1


def test_run_dataflow_add_column(tmp_path: Path) -> None:
    data = {"a": pa.array(range(100))}
    tbl = pa.Table.from_pydict(data)
    ckp_store = LanceCheckpointStore(str(tmp_path / "ckp"))

    tbl_path = tmp_path / "foo.lance"
    ds = lance.write_dataset(tbl, tbl_path)

    data = {"a": pa.array(range(100, 200))}
    tbl = pa.Table.from_pydict(data)
    ds = lance.write_dataset(tbl, tbl_path, mode="append")

    def return_none(batch: pa.RecordBatch) -> pa.RecordBatch:
        return pa.RecordBatch.from_pydict(
            {"b": pa.array([None] * batch.num_rows, pa.int32())}
        )

    new_frags = []
    new_schema = None
    for frag in ds.get_fragments():
        new_fragment, new_schema = frag.merge_columns(return_none, columns=["a"])
        new_frags.append(new_fragment)

    assert new_schema is not None
    merge = lance.LanceOperation.Merge(new_frags, new_schema)
    lance.LanceDataset.commit(tbl_path, merge, read_version=ds.version)

    run_dataflow_add_column(
        str(tbl_path),
        ["a"],
        {"b": add_one},
        ckp_store,
        DataflowOptions(
            runner="DirectRunner",
        ),
        batch_size=8,
        test_run=False,
    ).result.wait_until_finish()

    ds = lance.dataset(tbl_path)

    assert ds.to_table().to_pydict() == {
        "a": list(range(200)),
        "b": [x + 1 for x in range(200)],
    }


def test_dataflow_runner_test_run(tmp_path: Path) -> None:
    data = {"a": pa.array(range(100))}
    tbl = pa.Table.from_pydict(data)
    ckp_store = LanceCheckpointStore(str(tmp_path / "ckp"))

    tbl_path = tmp_path / "foo.lance"
    # do it twice to have two fragments
    ds = lance.write_dataset(tbl, tbl_path)

    data = {"a": pa.array(range(100, 200))}
    tbl = pa.Table.from_pydict(data)
    ds = lance.write_dataset(tbl, tbl_path, mode="append")

    def return_none(batch: pa.RecordBatch) -> pa.RecordBatch:
        return pa.RecordBatch.from_pydict(
            {"b": pa.array([None] * batch.num_rows, pa.int32())}
        )

    new_frags = []
    new_schema = None
    for frag in ds.get_fragments():
        new_fragment, new_schema = frag.merge_columns(return_none, columns=["a"])
        new_frags.append(new_fragment)

    assert new_schema is not None
    merge = lance.LanceOperation.Merge(new_frags, new_schema)
    lance.LanceDataset.commit(tbl_path, merge, read_version=ds.version)

    run_dataflow_add_column(
        str(tbl_path),
        ["a"],
        {"b": add_one},
        ckp_store,
        DataflowOptions(
            runner="DirectRunner",
        ),
        batch_size=8,
        test_run=True,
    ).result.wait_until_finish()

    ds = lance.dataset(tbl_path)

    assert ds.to_table().to_pydict() == {
        "a": list(range(200)),
        "b": [x + 1 for x in range(100)] + [None] * 100,
    }



# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: Copyright The Geneva Authors

import os
from collections.abc import Iterable
from functools import cached_property
from pathlib import Path
from typing import TYPE_CHECKING, Optional
from urllib.parse import urlparse

import attrs
import lancedb
import pyarrow as pa
from lancedb import DBConnection
from lancedb.common import DATA, Credential
from lancedb.pydantic import LanceModel
from overrides import override

from geneva.checkpoint import CheckpointStore
from geneva.config import CONFIG_LOADER, ConfigBase
from geneva.remote.client import RestfulLanceDBClient

if TYPE_CHECKING:
    from flightsql import FlightSQLClient

    from geneva.job.client import JobClient
    from geneva.runners.dataflow import DataflowOptions
    from geneva.table import Table


class Connection(DBConnection):
    """Geneva Connection."""

    def __init__(
        self,
        uri: str,
        *,
        region: str = "us-east-1",
        api_key: Credential | None = None,
        host_override: str | None = None,
        storage_options: dict[str, str] | None = None,
        checkpoint_store: CheckpointStore | None = None,
        dataflow_options: Optional["DataflowOptions"] = None,
        **kwargs,
    ) -> None:
        super().__init__()
        self._uri = uri
        self._region = region
        self._api_key = api_key
        self._host_override = host_override
        self._storage_options = storage_options
        self._ldb: DBConnection | None = None
        self._checkpoint_store = checkpoint_store

        self._flight_client: FlightSQLClient | None = None

        # Dataflow Options
        self._dataflow_options = dataflow_options

    def __repr__(self) -> str:
        return f"<LanceLake uri={self.uri}>"

    def __getstate__(self) -> dict:
        return {
            "uri": self._uri,
            "api_key": self._api_key,
            "host_override": self._host_override,
            "storage_options": self._storage_options,
            "region": self._region,
        }

    def __setstate__(self, state) -> None:
        self.__init__(state.pop("uri"), **state)

    def __enter__(self) -> "Connection":
        return self

    def __exit__(self, exc_type, exc_value, traceback) -> None:
        self.close()
        return None  # Don't suppress exceptions

    def close(self) -> None:
        """Close the connection."""
        if self._flight_client is not None:
            self._flight_client.close()

    @cached_property
    def _connect(self) -> DBConnection:
        """Returns the underlying lancedb connection."""
        if self._ldb is None:
            self._ldb = lancedb.connect(
                self.uri,
                region=self._region,
                api_key=self._api_key,
                host_override=self._host_override,
                storage_options=self._storage_options,
            )
        return self._ldb

    @cached_property
    def flight_client(self) -> "flightsql.FlightSQLClient":  # noqa: F821
        from flightsql import FlightSQLClient

        if self._flight_client is not None:
            return self._flight_client
        url = urlparse(self._host_override)
        hostname = url.hostname
        client = FlightSQLClient(
            host=hostname,
            port=10025,
            token="DATABASE_TOKEN",  # Dummy auth, not plugged in yet
            metadata={"database": self.uri},  # Name of the project-id
            features={"metadata-reflection": "true"},
            insecure=True,  # or False, up to you
        )
        self._flight_client = client
        return client

    @cached_property
    def _client(self) -> RestfulLanceDBClient:
        if (self._api_key is None) or (self._region is None):
            raise ValueError("API Key and Region must be provided.")

        return RestfulLanceDBClient(
            db_name=self._uri.removeprefix("db://"),
            region=self._region,
            api_key=self._api_key,
            host_override=self._host_override,
        )

    @override
    def table_names(
        self,
        page_token: str | None = None,
        limit: int | None = None,
    ) -> Iterable[str]:
        """List all available tables and views."""
        return self._connect.table_names(page_token=page_token, limit=limit or 10)

    @override
    def open_table(
        self,
        name: str,
        *,
        storage_options: dict[str, str] | None = None,
        index_cache_size: int | None = None,
    ) -> "Table":
        """Open a Lance Table.

        Parameters
        ----------
        name: str
            Name of the table.
        storage_options: dict[str, str], optional
            Additional options for the storage backend.
            Options already set on the connection will be inherited by the table,
            but can be overridden here. See available options at
            [https://lancedb.github.io/lancedb/guides/storage/](https://lancedb.github.io/lancedb/guides/storage/)


        """
        from .table import Table

        storage_options = storage_options or self._storage_options

        return Table(
            self,
            name,
            index_cache_size=index_cache_size,
            storage_options=storage_options,
        )

    @override
    def create_table(  # type: ignore
        self,
        name: str,
        data: DATA | None = None,
        schema: pa.Schema | LanceModel | None = None,
        mode: str = "create",
        exist_ok: bool = False,
        on_bad_vectors: str = "error",
        fill_value: float = 0.0,
        *,
        storage_options: dict[str, str] | None = None,
        **kwargs,
    ) -> "Table":  # type: ignore
        """Create a Table in the lake

        Parameters
        ----------
        name: str
            The name of the table
        data: The data to initialize the table, *optional*
            User must provide at least one of `data` or `schema`.
            Acceptable types are:

            - list-of-dict
            - pandas.DataFrame
            - pyarrow.Table or pyarrow.RecordBatch
        schema: The schema of the table, *optional*
            Acceptable types are:

            - pyarrow.Schema
            - [LanceModel][lancedb.pydantic.LanceModel]
        mode: str; default "create"
            The mode to use when creating the table.
            Can be either "create" or "overwrite".
            By default, if the table already exists, an exception is raised.
            If you want to overwrite the table, use mode="overwrite".
        exist_ok: bool, default False
            If a table by the same name already exists, then raise an exception
            if exist_ok=False. If exist_ok=True, then open the existing table;
            it will not add the provided data but will validate against any
            schema that's specified.
        on_bad_vectors: str, default "error"
            What to do if any of the vectors are not the same size or contain NaNs.
            One of "error", "drop", "fill".
        """
        from .table import Table

        self._connect.create_table(
            name,
            data,
            schema,
            mode,
            exist_ok=exist_ok,
            on_bad_vectors=on_bad_vectors,
            fill_value=fill_value,
            storage_options=storage_options,
            **kwargs,
        )
        return Table(self, name, storage_options=storage_options)

    def create_view(
        self,
        name: str,
        query: str,
        materialized: bool = False,
    ) -> "Table":
        """Create a View from a Query.

        Parameters
        ----------
        name: str
            Name of the view.
        query: Query
            Query to create the view.
        materialized: bool, optional
            If True, the view is materialized.
        """
        self.sql("CREATE VIEW {name} AS ({query})")
        return self.open_table(name)

    def drop_view(self, name: str) -> pa.Table:
        """Drop a view."""
        return self.sql(f"DROP VIEW {name}")

    @cached_property
    def jobs(self) -> "JobClient":
        """Geneva Jobs API

        Example
        -------

            # List all jobs
            >>> conn = connect("db://mydb")
            >>> jobs = conn.jobs.list(table="mytable",
                limit=500,
                filter="created_at > '2021-01-01'")

            # Start a new job
            >>> conn.jobs.start(table="mytable", column="virtual_col")
        """
        from geneva.job.client import JobClient

        return JobClient(rest_client=self._client)

    def sql(self, query: str) -> pa.Table:
        """Execute a raw SQL query.

        It uses the Flight SQL engine to execute the query.

        Parameters
        ----------
        query: str
            SQL query to execute

        Returns
        -------
        pyarrow.Table
            Result of the query in a `pyarrow.Table`

        TODO
        ----
        - Support pagination
        - Support query parameters
        """
        info = self.flight_client.execute(query)
        return self.flight_client.do_get(info.endpoints[0].ticket).read_all()


@attrs.define
class _GenavaConnectionConfig(ConfigBase):
    region: str = attrs.field(default="us-east-1")
    api_key: str | None = attrs.field(default=None)
    host_override: str | None = attrs.field(default=None)
    checkpoint: str | None = attrs.field(default=None)

    @classmethod
    @override
    def name(cls) -> str:
        return "connection"


def connect(
    uri: str | Path,
    *,
    region: str | None = None,
    api_key: Credential | str | None = None,
    host_override: str | None = None,
    storage_options: dict[str, str] | None = None,
    checkpoint: str | CheckpointStore | None = None,
    dataflow_options: Optional["DataflowOptions"] = None,
    **kwargs,
) -> Connection:
    """Connect to Geneva.

    Example
    -------

        >>> import geneva
        >>> conn = geneva.connect("db://my_dataset")
        >>> tbl = conn.open_table("youtube_dataset")

    Parameters
    ----------
    uri: geneva URI, or Path

    checkpoint: str or CheckpointStore, optional
        Checkpoint store for the connection. If not provided, will try to read from
        envvar "LANCE_CHECKPOINT", or from a configuration file.
    dataflow_options : DataflowOptions, optional
        If use dataflow engine, specify the options here.

    Returns
    -------
    Connection - A db connection

    """

    # load values from config if not provided via arguments
    config = CONFIG_LOADER.load(_GenavaConnectionConfig)
    region = region or config.region
    api_key = api_key or config.api_key
    api_key = Credential(api_key) if isinstance(api_key, str) else api_key
    host_override = host_override or config.host_override

    if checkpoint is None:
        checkpoint = os.getenv("LANCE_CHECKPOINT", "memory")
    if isinstance(checkpoint, str):
        checkpoint_store = CheckpointStore.from_uri(checkpoint)
    else:
        checkpoint_store = checkpoint
    return Connection(
        str(uri),
        region=region,
        api_key=api_key,
        host_override=host_override,
        storage_options=storage_options,
        checkpoint_store=checkpoint_store,
        dataflow_options=dataflow_options,
        **kwargs,
    )



# type: ignore

# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: Copyright The Geneva Authors

# This module contains an adapter that will close connections if they have not been
# used before a certain timeout. This is necessary because some load balancers will
# close connections after a certain amount of time, but the request module may not yet
# have received the FIN/ACK and will try to reuse the connection.
#
# TODO some of the code here can be simplified if/when this PR is merged:
# https://github.com/urllib3/urllib3/pull/3275


import datetime
import logging
import os

from requests.adapters import HTTPAdapter
from urllib3.connection import HTTPSConnection
from urllib3.connectionpool import HTTPSConnectionPool
from urllib3.poolmanager import PoolManager


def get_client_connection_timeout() -> int:
    return int(os.environ.get("LANCE_CLIENT_CONNECTION_TIMEOUT", "300"))


class LanceDBHTTPSConnection(HTTPSConnection):
    """
    HTTPSConnection that tracks the last time it was used.
    """

    idle_timeout: datetime.timedelta
    last_activity: datetime.datetime

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.last_activity = datetime.datetime.now()

    def request(self, *args, **kwargs) -> None:
        self.last_activity = datetime.datetime.now()
        super().request(*args, **kwargs)

    def is_expired(self) -> bool:
        return datetime.datetime.now() - self.last_activity > self.idle_timeout


def LanceDBHTTPSConnectionPoolFactory(client_idle_timeout: int):  # noqa: ANN201, N802
    """
    Creates a connection pool class that can be used to close idle connections.
    """

    class LanceDBHTTPSConnectionPool(HTTPSConnectionPool):
        # override the connection class
        ConnectionCls = LanceDBHTTPSConnection

        def __init__(self, *args, **kwargs) -> None:
            super().__init__(*args, **kwargs)

        def _get_conn(self, timeout: float | None = None):  # noqa: ANN202
            logging.debug("Getting https connection")
            conn = super()._get_conn(timeout)
            if conn.is_expired():
                logging.debug("Closing expired connection")
                conn.close()

            return conn

        def _new_conn(self):  # noqa: ANN202
            conn = super()._new_conn()
            conn.idle_timeout = datetime.timedelta(seconds=client_idle_timeout)
            return conn

    return LanceDBHTTPSConnectionPool


class LanceDBClientPoolManager(PoolManager):
    def __init__(
        self, client_idle_timeout: int, num_pools: int, maxsize: int, **kwargs
    ) -> None:
        super().__init__(num_pools=num_pools, maxsize=maxsize, **kwargs)
        # inject our connection pool impl
        connection_pool_class = LanceDBHTTPSConnectionPoolFactory(
            client_idle_timeout=client_idle_timeout
        )
        self.pool_classes_by_scheme["https"] = connection_pool_class


def LanceDBClientHTTPAdapterFactory():  # noqa: ANN201, N802
    """
    Creates an HTTPAdapter class that can be used to close idle connections
    """

    # closure over the timeout
    client_idle_timeout = get_client_connection_timeout()

    class LanceDBClientRequestHTTPAdapter(HTTPAdapter):
        def init_poolmanager(self, connections, maxsize, block=False) -> None:
            # inject our pool manager impl
            self.poolmanager = LanceDBClientPoolManager(
                client_idle_timeout=client_idle_timeout,
                num_pools=connections,
                maxsize=maxsize,
                block=block,
            )

    return LanceDBClientRequestHTTPAdapter



# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: Copyright The Geneva Authors

# a zip packager for local workspace

import hashlib
import logging
import re
import site
import sys
import tempfile
import zipfile
from pathlib import Path

import attrs

from geneva.config import CONFIG_LOADER, ConfigBase

_LOG = logging.getLogger(__name__)


@attrs.define
class _ZipperConfig(ConfigBase):
    output_path: Path | None = attrs.field(
        validator=attrs.validators.instance_of(Path),
        converter=attrs.converters.optional(Path),
    )

    @classmethod
    def name(cls) -> str:
        return "zipper"


@attrs.define
class WorkspaceZipper:
    path: Path = attrs.field(
        converter=attrs.converters.pipe(
            Path,
            Path.resolve,
            Path.absolute,
        )
    )

    @path.validator
    def _path_validator(self, attribute, value: Path) -> None:
        if not value.is_dir():
            raise ValueError("path must be a directory")

        # make sure the path is the current working directory, or
        # is part of sys.path
        if value == Path.cwd().resolve().absolute():
            return

        sys_paths = {Path(x).resolve().absolute() for x in sys.path}

        if value not in sys_paths:
            raise ValueError("path must be cwd or part of sys.path")

    ouput_dir: Path = attrs.field()

    @ouput_dir.default
    def _output_dir_default(self) -> Path:
        config = CONFIG_LOADER.load(_ZipperConfig)
        if config.output_path is not None:
            return config.output_path
        return self.path / ".geneva"

    ignore_regexs: list[re.Pattern] = attrs.field(
        factory=list,
        converter=lambda x: [re.compile(r) for r in x],
    )

    def zip(self) -> tuple[Path, str]:
        """
        create a zip file for the workspace

        return the path of the zip file and the sha256 hash of the zip file
        """
        zip_path = self.ouput_dir / "workspace.zip"
        with zipfile.ZipFile(zip_path, "w", compression=zipfile.ZIP_DEFLATED) as z:
            for child in self.path.rglob("*"):
                if any(r.match(child) for r in self.ignore_regexs):
                    continue
                arcname = child.relative_to(self.path)
                z.write(child, arcname.as_posix())
        return zip_path, hashlib.sha256(zip_path.read_bytes()).hexdigest()


@attrs.define
class _UnzipperConfig(ConfigBase):
    output_dir: Path | None = attrs.field(
        converter=attrs.converters.optional(Path),
        default=None,
    )

    @classmethod
    def name(cls) -> str:
        return "unzipper"


# Kept separate from the zipper to avoid config mess
@attrs.define
class WorkspaceUnzipper:
    output_dir: Path = attrs.field(
        converter=attrs.converters.pipe(
            attrs.converters.default_if_none(
                CONFIG_LOADER.load(_UnzipperConfig).output_dir,
            ),
            attrs.converters.default_if_none(
                factory=tempfile.mkdtemp,
            ),
            Path,
            Path.resolve,
            Path.absolute,
        ),
        default=None,
    )

    def unzip(self, zip_path: Path, *, checksum: str) -> None:
        """
        extract the zip file to the workspace
        """
        if hashlib.sha256(zip_path.read_bytes()).hexdigest() != checksum:
            raise ValueError("workspace zip checksum mismatch")

        with zipfile.ZipFile(zip_path, "r") as z:
            z.extractall(self.output_dir)
        _LOG.info("extracted workspace to %s", self.output_dir)

        site.addsitedir(self.output_dir.as_posix())
        _LOG.info("added %s to sys.path", self.output_dir)



# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: Copyright The Geneva Authors

import heapq
import logging
import re
import uuid
from collections.abc import Iterator
from typing import cast

import attrs
import lance
import pyarrow as pa
import ray.actor
import ray.data
from tqdm.auto import tqdm

from geneva.apply import LanceRecordBatchUDFApplier, ReadTask, plan_read
from geneva.checkpoint import CheckpointStore
from geneva.debug.logger import CheckpointStoreErrorLogger
from geneva.transformer import UDF

_LOG = logging.getLogger(__name__)


def _combine_chunks(arr: pa.Array) -> pa.Array:
    if isinstance(arr, pa.ChunkedArray):
        assert len(arr.chunks) == 1
        return arr.chunks[0]
    return arr


def _write_fragment(
    uri: str,
    column_name: str,
    store: CheckpointStore,
    checkpoint_keys: Iterator[dict[str, str]],
) -> lance.fragment.DataFile:
    dataset = lance.dataset(uri)
    import more_itertools

    def _iter() -> Iterator[pa.RecordBatch]:
        for batch in checkpoint_keys:
            res = {
                key: _combine_chunks(store[value]["data"])
                for key, value in batch.items()
            }
            yield pa.RecordBatch.from_pydict(res)

    it = more_itertools.peekable(_iter())
    rbr = pa.RecordBatchReader.from_batches(it.peek().schema, it)

    # TODO: this doesn't support struct or complex schema yet
    new_data = lance.fragment.write_fragments(
        rbr,
        uri,
        max_rows_per_file=1 << 31,
        max_bytes_per_file=1 << 40,
    )

    assert len(new_data) == 1
    new_data = new_data[0]
    assert len(new_data.files) == 1
    new_datafile = new_data.files[0]

    # MASSIVE HACK: open up an API to get the field id from the column name
    field_id = re.compile(rf'name: "{column_name}", id: (?P<field_id>[\d]*),').findall(
        str(dataset.lance_schema)
    )
    assert len(field_id) == 1
    field_id = int(field_id[0])

    new_datafile.fields = [field_id]
    new_datafile.column_indices = [0]

    return new_datafile


@ray.remote
@attrs.define
class ApplierActor:
    applier: LanceRecordBatchUDFApplier

    def run(self, task) -> tuple[ReadTask, dict[str, str]]:
        return task, self.applier.run(task)


ApplierActor: ray.actor.ActorClass = cast(ray.actor.ActorClass, ApplierActor)


@ray.remote
@attrs.define
class FragementWriter:
    uri: str
    column_name: str
    store: CheckpointStore

    frags: list = attrs.field(factory=list)

    written: bool = attrs.field(init=False, default=False)

    def write(self) -> lance.fragment.DataFile:
        if self.written:
            raise RuntimeError("Already written")

        def _key_iter() -> Iterator[dict[str, str]]:
            while self.frags:
                _, item = heapq.heappop(self.frags)
                yield item

        file = _write_fragment(self.uri, self.column_name, self.store, _key_iter())
        self.written = True
        return file

    def add_task_result(self, task: ReadTask, result) -> None:
        if self.written:
            raise RuntimeError("Already written")

        heapq.heappush(self.frags, (task.offset, result))


FragementWriter: ray.actor.ActorClass = cast(ray.actor.ActorClass, FragementWriter)


def run_ray_add_column(
    uri: str,
    columns: list[str],
    transforms: dict[str, UDF],
    checkpoint_store: CheckpointStore,
    /,
    job_id: str | None = None,
    batch_size: int = 8,
    read_version: int | None = None,
    concurrency: int = 8,
    plan_shuffle_buffer_size: int = 1024,
    test_run: bool = True,
    **kwargs,
) -> None:
    if read_version is None:
        read_version = lance.dataset(uri).version

    job_id = job_id or uuid.uuid4().hex

    applier = LanceRecordBatchUDFApplier(
        udfs=transforms,
        checkpoint_store=checkpoint_store,
        error_logger=CheckpointStoreErrorLogger(job_id, checkpoint_store),
    )

    pool = ray.util.ActorPool(
        [ApplierActor.remote(applier=applier) for _ in range(concurrency)]
    )

    plan = plan_read(
        uri,
        columns,
        batch_size=batch_size,
        read_version=read_version,
        shuffle_buffer_size=plan_shuffle_buffer_size,
    )

    applier_iter = pool.map_unordered(
        lambda actor, value: actor.run.remote(value),
        # the API says list, but iterables are fine
        plan,
    )

    ds = lance.dataset(uri, version=read_version)
    writers = {
        frag.fragment_id: FragementWriter.remote(
            uri, list(transforms.keys())[0], checkpoint_store
        )
        for frag in ds.get_fragments()
    }

    applier_pbar = tqdm(applier_iter, total=len(plan))
    applier_pbar.set_description("Applying UDFs")

    futs = []
    for item in applier_pbar:
        task: ReadTask = item[0]
        result = item[1]

        frag_id = task.frag_id
        writer = writers[frag_id]
        futs.append(writer.add_task_result.remote(task, result))
    ray.wait(futs)

    writers = {frag_id: writer.write.remote() for frag_id, writer in writers.items()}

    _LOG.info("Committing %d fragments", len(writers))

    writer_pbar = tqdm(writers.items(), total=len(writers))
    writer_pbar.set_description("Writing Fragments")

    operation = lance.LanceOperation.DataReplacement(
        replacements=[
            lance.LanceOperation.DataReplacementGroup(
                fragment_id=frag_id,
                new_file=ray.get(writer),
            )
            for frag_id, writer in writer_pbar
        ]
    )

    lance.LanceDataset.commit(uri, operation, read_version=read_version)



# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: Copyright The Geneva Authors

from geneva.runners.dataflow.dataflow import (
    DataflowOptions,
    run_dataflow_add_column,
)

__all__ = ["run_dataflow_add_column", "DataflowOptions"]



# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: Copyright The Geneva Authors

#!/usr/bin/env python3

import argparse
import logging

from geneva.debug.entrypoint import register_debug_parser
from geneva.docker.packager import register_packager_parser
from geneva.tui import register_console_parser


def main() -> None:
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-v",
        "--verbose",
        action="store_true",
        default=False,
        help="Enable verbose logging",
    )

    subparsers = parser.add_subparsers(dest="actions", required=True)

    register_debug_parser(subparsers)
    register_packager_parser(subparsers)
    register_console_parser(subparsers)

    args = parser.parse_args()
    if args.verbose:
        logging.getLogger().setLevel(logging.DEBUG)
    else:
        logging.getLogger().setLevel(logging.INFO)
    args.func(args)


if __name__ == "__main__":
    main()
